from __future__ import annotations

import contextlib
import dataclasses as dc
import logging
import shutil
import tempfile
from pathlib import Path
from typing import Generator

from .run1 import EMode, OMode, Paths, mkpaths, runc

logger = logging.getLogger(__name__)


@dc.dataclass
class Runner:
    verbose: bool
    dryrun: bool | None = None
    exe: Paths | None = None
    cwd: Path | None = None
    overrides: dict[str, str] | None = None
    log: logging.Logger | None = None

    @staticmethod
    @contextlib.contextmanager
    def tmpdir(source: Path | None = None) -> Generator[Path, None, None]:
        wdir = source if source else Path(tempfile.mkdtemp())
        wdir.mkdir(parents=True, exist_ok=True)
        try:
            yield wdir
        finally:
            if not source:
                shutil.rmtree(wdir, ignore_errors=True)

    def __call__(
        self,
        args: Paths,
        capture: bool = False,
        verbose: bool | None = None,
        dryrun: bool | None = None,
        cwd: Path | str | bool | None = None,
        overrides: dict[str, str] | None = None,
        log: logging.Logger | None = None,
    ) -> str | bytes | None:
        # capture/verbose are interacting to control how the stdout is handled
        check = (capture, self.verbose if verbose is None else verbose)
        stdout: OMode = "null"
        if check == (True, False):
            stdout = "capture"
        elif check == (True, True):
            stdout = "capture+display"
        elif check == (False, True):
            stdout = "display"
        elif check == (False, False):
            stdout = "null"
        else:
            raise RuntimeError(f"un-handled value {check=}")
        stderr: EMode = "display" if self.verbose else "null"

        dryrun = self.dryrun if dryrun is None else dryrun
        if "capture" in stdout and dryrun:
            raise RuntimeError("cannot dryrun and caputure")

        cwd = cwd or self.cwd
        overrides = overrides or self.overrides
        log = log or self.log or logger

        fullargs = mkpaths(args)
        if self.exe:
            fullargs = [*mkpaths(self.exe), *fullargs]

        log.debug("%srun: %s", "(dry-run) " if dryrun else "", " ".join(fullargs))
        if dryrun:
            return None
        return runc(fullargs, stdout=stdout, stderr=stderr, overrides=overrides, cwd=cwd)


if __name__ == "__main__":
    runner = Runner(True)
    y = runner(["ls", "-l"], capture=False, verbose=True)
    print(y)

    # --git-dir=$dest/.git --work-tree $dest
    runner = Runner(True, exe=["git", "--git-dir", "{workdir}/.git"])
    runner("status")



# from https://packaging.python.org/en/latest/specifications/inline-script-metadata/#inline-script-metadata
import re
from pathlib import Path

import tomllib

REGEX = r"(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)+)^# ///$"


def read_header(path: Path | str) -> dict | None:
    text = path if isinstance(path, str) else path.read_text()

    name = "script"
    matches = list(filter(lambda m: m.group("type") == name, re.finditer(REGEX, text)))
    if len(matches) > 1:
        raise ValueError(f"Multiple {name} blocks found")
    elif len(matches) == 1:
        content = "".join(line[2:] if line.startswith("# ") else line[1:] for line in matches[0].group("content").splitlines(keepends=True))
        return tomllib.loads(content)
    else:
        return None


if __name__ == "__main__":
    print(read_header(Path("support/builder.py")))



from acbox import utils


def test_load_mod(resolver):
    path = resolver.lookup("simple-script.py")

    mod = utils.loadmod(path)
    assert callable(mod.hello)


def test_load_remote():
    mod = utils.loadmod("https://raw.githubusercontent.com/cav71/acbox/refs/heads/main/tests/test_utils.py")
    assert hasattr(mod, "test_load_remote")


def test_diffdict():
    left = {
        "a": 1,
        "b": 2,
    }
    right = {
        "b": 2,
        "c": 3,
    }
    assert utils.diffdict(left, right, na=utils.NA) == {"a": (1, utils.NA), "c": (utils.NA, 3)}
    assert utils.diffdict(left, right, na="N/A") == {"a": (1, "N/A"), "c": ("N/A", 3)}
    assert utils.diffdict(left, right, {"a"}, na="N/A") == {"c": ("N/A", 3)}



import argparse
import logging
import sys
import types

from .base import ArgumentParserBase

log = logging.getLogger(__name__)


def log_sys_info(modules: list[types.ModuleType]):
    log.debug("interpreter: %s", sys.executable)


class ArgumentParser(ArgumentParserBase):
    def parse_args(self, args=None, namespace=None):
        options = super().parse_args(args, namespace)

        # reserver attributes
        for reserved in [
            "modules",
            "error",
        ]:
            if not hasattr(options, reserved):
                continue
            raise RuntimeError(f"cannot add an argument with dest='{reserved}'")
        options.error = self.error
        options.modules = self.modules

        for callback in self.callbacks:
            if not callback:
                continue
            options = callback(options) or options

        log_sys_info(self.modules)
        return options

    @classmethod
    def get_parser(cls, modules: list[types.ModuleType], **kwargs):
        class Formatter(
            argparse.RawTextHelpFormatter,
            argparse.RawDescriptionHelpFormatter,
            argparse.ArgumentDefaultsHelpFormatter,
        ):
            pass

        return cls(modules, formatter_class=Formatter, **kwargs)



from __future__ import annotations

import contextlib
import os
import shutil
import tempfile
from pathlib import Path
from typing import Callable, Generator


def which_n(exe: str | Path) -> list[Path] | None:
    candidates: list[Path] | None = None
    for srcdir in os.environ.get("PATH", "").split(os.pathsep):
        for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
            path = srcdir / Path(exe).with_suffix(ext)
            if not path.exists():
                continue
            if candidates is None:
                candidates = []
            candidates.append(path)
    return candidates


def which(exe: str | Path) -> Path | None:
    candidates = which_n(exe)
    if candidates is None:
        return None
    return candidates[0]


@contextlib.contextmanager
def tmpdir(source: Path | None) -> Generator[Path, None, None]:
    wdir = source if source else Path(tempfile.mkdtemp())
    wdir.mkdir(parents=True, exist_ok=True)
    try:
        yield wdir
    finally:
        if not source:
            shutil.rmtree(wdir, ignore_errors=True)


@contextlib.contextmanager
def backups() -> Generator[Callable[[Path | str], tuple[Path, Path]], None, None]:
    pathlist: list[Path] = []

    def save(path: Path | str) -> tuple[Path, Path]:
        nonlocal pathlist
        original = Path(path).expanduser().absolute()
        backup = original.parent / f"{original.name}.bak"
        if backup.exists():
            raise RuntimeError("backup file present", backup)
        shutil.copy(original, backup)
        pathlist.append(backup)
        return original, backup

    try:
        yield save
    finally:
        for backup in pathlist:
            original = backup.with_suffix("")
            original.unlink()
            shutil.move(backup, original)



from __future__ import annotations

from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType
from typing import Any
from urllib.parse import urlparse
from urllib.request import urlopen


def loadmod(path: Path | str) -> ModuleType:
    txt = None
    if (parsed := urlparse(str(path))).scheme in {"http", "https"}:
        txt = str(urlopen(str(path)).read(), encoding="utf-8")
    elif parsed.scheme in {"file"}:
        txt = Path(parsed.netloc).read_text()

    if txt is not None:
        mod = ModuleType(str(path).rpartition("/")[2])
        exec(txt, mod.__dict__)
        return mod

    spec = spec_from_file_location(Path(path).name, Path(path))
    module = module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore
    return module


class NA:
    pass


def diffdict(
    left: dict[str, Any], right: dict[str, Any], exclude: list[str] | None = None, na: str | type[NA] = NA
) -> dict[str, tuple[Any, Any]]:
    result = {}
    for key in sorted(set(left) | set(right)):
        if exclude and key in exclude:
            continue
        if key not in left:
            result[key] = (na, right[key])
        elif key not in right:
            result[key] = (left[key], na)
        elif left[key] != right[key]:
            result[key] = (left[key], right[key])
    return result



from argparse import Namespace

import click
from click.testing import CliRunner

from acbox import clickwrapper


def test_base():
    @click.command()
    @clickwrapper.clickwrap()
    def main(args: Namespace) -> None:
        pass

    runner = CliRunner()
    result = runner.invoke(main, ["--help"])
    assert result.exit_code == 0
    assert (
        """
Usage: main [OPTIONS]

Options:
  -q, --quiet
  -v, --verbose
  --help         Show this message and exit.
""".lstrip()
        == result.stdout
    )



import shutil
import tempfile
from pathlib import Path


def process(txt: str, defines: dict[str, str]) -> str:
    result = []
    for line in txt.split("\n"):
        newline = line
        for key, value in defines.items():
            newline = newline.replace(f"@{key}@", value)
        result.append(newline)
    return "\n".join(result)


def fix(path: Path, defines: dict[str, str], inplace: bool) -> str:
    with tempfile.NamedTemporaryFile(mode="w+") as fp:
        shutil.copymode(path, fp.name)
        fp.write(process(path.read_text(), defines))
        fp.flush()
        fp.seek(0)
        if inplace:
            shutil.copy(fp.name, path)
        return fp.read()
